@@ -1,4 +1,5 @@
 import hashlib
+import logging
 import math
 import os.path
 import re
@@ -7,12 +8,13 @@ from typing import Any, Dict, Iterable, List, Mapping, Optional, Sequence, Tuple
 
 import pandas as pd
 
-from countess.utils.pandas import get_all_columns
-
 PARAM_DIGEST_HASH = "sha256"
 
 
-def make_prefix_groups(strings: List[str]) -> Dict[str, List[str]]:
+logger = logging.getLogger(__name__)
+
+
+def make_prefix_groups(strings: Iterable[str]) -> Dict[str, List[str]]:
     groups: Dict[str, List[str]] = {}
     for s in strings:
         if m := re.match(r"(.*?_+)([^_]+)$", s):
@@ -48,7 +50,7 @@ class BaseParam:
     def get_hash_value(self):
         raise NotImplementedError(f"Implement {self.__class__.__name__}.get_hash_value()")
 
-    def set_column_choices(self, choices):
+    def set_column_choices(self, choices: Mapping[str, bool]):
         pass
 
 
@@ -163,6 +165,9 @@ class NumericParam(ScalarWithOperatorsParam):
 
     def set_value(self, value):
         try:
+            if type(value) is str:
+                if m := re.match(r"(\d+(\.\d*)?)", value):
+                    value = m.group(1)
             self._value = self.var_type(value)
         except (TypeError, ValueError):
             self.reset_value()
@@ -277,23 +282,50 @@ class StringCharacterSetParam(StringParam):
         return self.__class__(self.label, self.value, character_set=self.character_set)
 
 
-class FileParam(StringParam):
-    """A StringParam for holding a filename."""
+class FileBaseParam(StringParam):
+    """A StringParam for holding a filename, either to be read or written."""
 
     file_types = [("Any", "*")]
 
-    _hash = None
-
-    def __init__(self, label: str, value=None, file_types=None):
+    def __init__(self, label: str, value=None, file_types=None, base_dir="."):
         super().__init__(label, value)
         if file_types is not None:
             self.file_types = file_types
+        self.base_dir = base_dir
+
+    def get_file_path(self):
+        return os.path.abspath(os.path.join(self.base_dir, self.value))
+
+    def set_base_dir(self, base_dir):
+        if self.value:
+            self.value = os.path.relpath(self.get_file_path(), base_dir)
+        self.base_dir = base_dir
+
+    def get_parameters(self, key, base_dir="."):
+        if self.value:
+            if base_dir:
+                path = os.path.relpath(self.get_file_path(), base_dir)
+            else:
+                path = os.path.abspath(self.get_file_path())
+        else:
+            path = None
+
+        return [(key, path)]
+
+    def copy(self) -> "FileBaseParam":
+        return self.__class__(self.label, self.value, file_types=self.file_types, base_dir=self.base_dir)
+
+
+class FileParam(FileBaseParam):
+    """A filename to be read as input."""
+
+    _hash = None
 
     def get_file_hash(self):
         if not self.value:
             return "0"
         try:
-            with open(self.value, "rb") as file:
+            with open(self.get_file_path(), "rb") as file:
                 try:
                     # Python 3.11
                     digest = hashlib.file_digest(file, PARAM_DIGEST_HASH)
@@ -308,20 +340,6 @@ class FileParam(StringParam):
         except IOError:
             return "0"
 
-    def get_parameters(self, key, base_dir="."):
-        if self.value:
-            if base_dir:
-                path = os.path.relpath(self.value, base_dir)
-            else:
-                path = os.path.abspath(self.value)
-        else:
-            path = None
-
-        return [(key, path)]
-
-    def copy(self) -> "FileParam":
-        return self.__class__(self.label, self.value, file_types=self.file_types)
-
     def get_hash_value(self) -> str:
         # For reproducability, we don't actually care about the filename, just
         # its hash.
@@ -330,13 +348,8 @@ class FileParam(StringParam):
         return self._hash
 
 
-class FileSaveParam(StringParam):
-    file_types = [("Any", "*")]
-
-    def __init__(self, label: str, value=None, file_types=None):
-        super().__init__(label, value)
-        if file_types is not None:
-            self.file_types = file_types
+class FileSaveParam(FileBaseParam):
+    """A filename to save output as."""
 
 
 class DictChoiceParam(ScalarWithOperatorsParam):
@@ -471,10 +484,10 @@ class ChoiceParam(ScalarWithOperatorsParam):
 
 class DataTypeChoiceParam(ChoiceParam):
     DATA_TYPES: Mapping[str, tuple[type, Any, Type[ScalarParam]]] = {
-        "string": (str, "", StringParam),
-        "number": (float, math.nan, FloatParam),
-        "integer": (int, 0, IntegerParam),
-        "boolean": (bool, False, BooleanParam),
+        "STRING": (str, "", StringParam),
+        "FLOAT": (float, math.nan, FloatParam),
+        "INTEGER": (int, 0, IntegerParam),
+        "BOOLEAN": (bool, False, BooleanParam),
     }
 
     def __init__(self, label: str, value: Optional[str] = None, choices: Optional[Iterable[str]] = None):
@@ -489,6 +502,8 @@ class DataTypeChoiceParam(ChoiceParam):
             return None
 
     def cast_value(self, value):
+        if self.value not in self.DATA_TYPES:
+            return None
         if value is not None:
             try:
                 return self.DATA_TYPES[self.value][0](value)
@@ -542,13 +557,18 @@ class ColumnChoiceParam(ChoiceParam):
     """A ChoiceParam which DaskTransformPlugin knows
     it should automatically update with a list of columns"""
 
-    def set_column_choices(self, choices):
-        self.set_choices(list(choices))
+    def set_column_choices(self, choices: Mapping[str, bool]):
+        self.set_choices(list(choices.keys()))
 
     def get_column(self, df):
         return _dataframe_get_column(df, self.value)
 
 
+class NumericColumnChoiceParam(ColumnChoiceParam):
+    def set_column_choices(self, choices: Mapping[str, bool]):
+        self.set_choices([c for c, n in choices.items() if n])
+
+
 class ColumnOrNoneChoiceParam(ColumnChoiceParam):
     DEFAULT_VALUE = "— NONE —"
 
@@ -569,16 +589,15 @@ class ColumnOrNoneChoiceParam(ColumnChoiceParam):
 
 
 class ColumnGroupChoiceParam(ChoiceParam):
-    def set_column_choices(self, choices):
-        self.set_choices([n + "*" for n in make_prefix_groups(choices).keys()])
+    def set_column_choices(self, choices: Mapping[str, bool]):
+        self.set_choices([n + "*" for n in make_prefix_groups(choices.keys()).keys()])
 
     def get_column_prefix(self):
         return self.value.removesuffix("*")
 
     def get_column_names(self, df):
         prefix = self.get_column_prefix()
-        column_names = get_all_columns(df).keys()
-        return [n for n in column_names if n.startswith(prefix)]
+        return [n for n in df.columns if n.startswith(prefix)]
 
     def get_column_suffixes(self, df):
         prefix = self.get_column_prefix()
@@ -602,6 +621,11 @@ class ColumnGroupOrNoneChoiceParam(ColumnGroupChoiceParam):
             return None
         return super().get_column_prefix()
 
+    def get_column_names(self, df):
+        if self.is_none():
+            return []
+        return super().get_column_names(df)
+
 
 class ColumnOrIndexChoiceParam(ColumnChoiceParam):
     DEFAULT_VALUE = "— INDEX —"
@@ -626,7 +650,7 @@ class ColumnOrStringParam(ColumnChoiceParam):
     DEFAULT_VALUE: Any = ""
     PREFIX = "— "
 
-    def set_column_choices(self, choices):
+    def set_column_choices(self, choices: Mapping[str, bool]):
         self.set_choices([self.PREFIX + c for c in choices])
 
     def get_column_name(self) -> Optional[str]:
@@ -634,18 +658,26 @@ class ColumnOrStringParam(ColumnChoiceParam):
             return self.value[len(self.PREFIX) :]
         return None
 
+    def get_string_value(self) -> Optional[str]:
+        if type(self.value) is str and not self.value.startswith(self.PREFIX):
+            return self.value
+        return None
+
     def get_value_from_dict(self, data: dict) -> str:
         if type(self.value) is str and self.value.startswith(self.PREFIX):
             return data[self.value[len(self.PREFIX) :]]
         else:
             return self.value
 
-    def get_column_or_value(self, df: pd.DataFrame, numeric: bool) -> Union[float, str, pd.Series]:
-        if type(self.value) is str and self.value.startswith(self.PREFIX):
-            col = df[self.value[len(self.PREFIX) :]]
-            return col.astype(float if numeric else str)
-        else:
-            return float(self.value) if numeric else str(self.value)
+    def get_column_or_value(self, df: pd.DataFrame, numeric: bool) -> Union[float, str, pd.Series, None]:
+        try:
+            if type(self.value) is str and self.value.startswith(self.PREFIX):
+                col = df[self.value[len(self.PREFIX) :]]
+                return col.astype(float if numeric else str)
+            else:
+                return float(self.value) if numeric else str(self.value)
+        except ValueError:
+            return None
 
     def set_choices(self, choices: Iterable[str]):
         self.choices = list(choices)
@@ -662,6 +694,9 @@ class ColumnOrStringParam(ColumnChoiceParam):
 class ColumnOrIntegerParam(ColumnOrStringParam):
     DEFAULT_VALUE: int = 0
 
+    def set_column_choices(self, choices: Mapping[str, bool]):
+        self.set_choices([self.PREFIX + c for c, n in choices.items() if n])
+
     def __init__(
         self,
         label: str,
@@ -671,6 +706,11 @@ class ColumnOrIntegerParam(ColumnOrStringParam):
         super().__init__(label, choices=choices)
         self.value = value
 
+    def get_integer_value(self) -> Optional[int]:
+        if type(self.value) is str and not self.value.startswith(self.PREFIX):
+            return int(self.value)
+        return None
+
     def clean_value(self, value):
         if type(value) is str and value.startswith(self.PREFIX):
             return value
@@ -727,14 +767,14 @@ class HasSubParametersMixin:
                 param.set_parameter(subkey, value, base_dir)
         else:
             param = self.params[key]
+            assert isinstance(param, ScalarParam)
+
             if isinstance(param, (FileParam, FileSaveParam)) and value is not None:
-                # XXX this is a clumsy approach?
-                param.set_value(os.path.join(base_dir, str(value)))
+                param.set_base_dir(base_dir)
 
-            elif isinstance(param, ScalarParam):
-                param.set_value(value)
+            param.set_value(value)
 
-    def set_column_choices(self, choices):
+    def set_column_choices(self, choices: Mapping[str, bool]):
         for p in self.params.values():
             p.set_column_choices(choices)
 
@@ -773,8 +813,7 @@ class ArrayParam(BaseParam):
         read_only: Optional[bool] = None,
     ):
         super().__init__(label)
-        self.param = param
-
+        self.param = param.copy()
         self.min_size = min_size
         self.max_size = max_size
         if read_only is not None:
@@ -835,7 +874,7 @@ class ArrayParam(BaseParam):
             digest.update(p.get_hash_value().encode("utf-8"))
         return digest.hexdigest()
 
-    def set_column_choices(self, choices):
+    def set_column_choices(self, choices: Mapping[str, bool]):
         self.param.set_column_choices(choices)
         for p in self.params:
             p.set_column_choices(choices)
@@ -876,10 +915,10 @@ class PerColumnArrayParam(ArrayParam):
             yield f"{key}.{n}._label", p.label
             yield from p.get_parameters(f"{key}.{n}", base_dir)
 
-    def set_column_choices(self, choices):
+    def set_column_choices(self, choices: Mapping[str, bool]):
         params_by_label = {p.label: p for p in self.params}
         self.params = []
-        for label in choices:
+        for label in choices.keys():
             if label in params_by_label:
                 self.params.append(params_by_label[label])
             else:
@@ -892,6 +931,11 @@ class PerColumnArrayParam(ArrayParam):
             yield p.label, p
 
 
+class PerNumericColumnArrayParam(PerColumnArrayParam):
+    def set_column_choices(self, choices: Mapping[str, bool]):
+        super().set_column_choices({k: True for k, v in choices.items() if v})
+
+
 class FileArrayParam(ArrayParam):
     """FileArrayParam is an ArrayParam arranged per-file.  Using this class
     really just marks it as expecting to be populated from an open file
