@@ -1,65 +1,32 @@
 import logging
 import re
+import secrets
+import threading
 import time
-from queue import Empty, Queue
-from threading import Thread
+from enum import Enum
 from typing import Any, Iterable, Optional
 
-from countess.core.plugins import BasePlugin, FileInputPlugin, ProcessPlugin, get_plugin_classes
+import duckdb
+
+from countess.core.plugins import BasePlugin, DuckdbPlugin, get_plugin_classes
+from countess.utils.duckdb import duckdb_source_to_view
 
 PRERUN_ROW_LIMIT = 100000
 
 logger = logging.getLogger(__name__)
 
 
-class SentinelQueue(Queue):
-
-    """This is an easy and a bit lazy way of making a queue iterable.
-    The writer is expected to call `queue.finish()` when it is done and
-    the reader can treat the queue like an iterable."""
-
-    # catch attempts to 'put' more data onto the queue after it has finished.
-    finished = False
-
-    # Handle multiple threads reading from the
-    # queue in parallel: once the sentinel has been received by any thread
-    # all further attempts to read get StopIterations.
-    stopped = False
-
-    class SENTINEL:
-        pass
-
-    def finish(self):
-        self.put(self.SENTINEL)
-        self.finished = True
-
-    def __iter__(self):
-        return self
-
-    def get_or_raise_on_stopped(self):
-        while True:
-            if self.stopped:
-                raise StopIteration
-            try:
-                return super().get(timeout=1)
-            except Empty:
-                pass
-
-    def __next__(self):
-        val = self.get_or_raise_on_stopped()
-        if val is self.SENTINEL:
-            self.stopped = True
-            raise StopIteration
-        return val
-
-    def put(self, item, block=True, timeout=None):
-        if self.finished:
-            raise ValueError("SentinelQueue stopped")
-        super().put(item, block, timeout)
+class PipelineNodeStatus(Enum):
+    INIT = 0
+    WAIT = 1
+    WORK = 2
+    DONE = 3
+    STOP = 4
 
 
 class PipelineNode:
     name: str
+    uuid: str
     plugin: Optional[BasePlugin] = None
     position: Optional[tuple[float, float]] = None
     sort_column: int = 0
@@ -67,21 +34,18 @@ class PipelineNode:
     notes: Optional[str] = None
     parent_nodes: set["PipelineNode"]
     child_nodes: set["PipelineNode"]
-    config: Optional[list[tuple[str, str, str]]] = None
-    result: Any = None
     is_dirty: bool = True
-
-    output_queues: set[SentinelQueue]
-    counter_in: int = 0
-    counter_out: int = 0
+    result: Any = None
 
     # XXX config is a cache for config loaded from the file
     # at config load time, if it is present it is loaded the
     # first time the plugin is prerun.
+    config: Optional[list[tuple[str, str, str]]] = None
 
     def __init__(
         self,
         name: str,
+        uuid: Optional[str] = None,
         plugin: Optional[BasePlugin] = None,
         position: Optional[tuple[float, float]] = None,
         notes: Optional[str] = None,
@@ -89,6 +53,7 @@ class PipelineNode:
         sort_descending: bool = False,
     ):
         self.name = name
+        self.uuid = uuid or secrets.token_hex(16)
         self.plugin = plugin
         self.position = position or (0.5, 0.5)
         self.sort_column = sort_column
@@ -96,7 +61,6 @@ class PipelineNode:
         self.notes = notes
         self.parent_nodes = set()
         self.child_nodes = set()
-        self.output_queues = set()
         self.config: list[tuple[str, str, str]] = []
 
     def set_config(self, key, value, base_dir):
@@ -111,85 +75,6 @@ class PipelineNode:
     def is_descendant_of(self, node: "PipelineNode") -> bool:
         return (self in node.child_nodes) or any((self.is_descendant_of(n) for n in node.child_nodes))
 
-    def add_output_queue(self) -> SentinelQueue:
-        queue = SentinelQueue(maxsize=3)
-        self.output_queues.add(queue)
-        return queue
-
-    def queue_output(self, result):
-        for data in result:
-            self.counter_out += 1
-            # XXX can we do this out-of-order if any queues are full?
-            for queue in self.output_queues:
-                queue.put(data)
-            logger.info("%s: %d/%d", self.name, self.counter_out, self.counter_in)
-
-    def finish_output(self):
-        for queue in self.output_queues:
-            queue.finish()
-
-    def run_multithread(self, queue: SentinelQueue, name: str, row_limit: Optional[int] = None):
-        assert isinstance(self.plugin, ProcessPlugin)
-        for data_in in queue:
-            self.counter_in += 1
-            self.plugin.preprocess(data_in, name)
-            self.queue_output(self.plugin.process(data_in, name))
-
-    def run_subthread(self, queue: SentinelQueue, name: str, row_limit: Optional[int] = None):
-        assert isinstance(self.plugin, ProcessPlugin)
-
-        for data_in in queue:
-            self.counter_in += 1
-            self.plugin.preprocess(data_in, name)
-            self.queue_output(self.plugin.process(data_in, name))
-        self.queue_output(self.plugin.finished(name))
-
-    def run_thread(self, row_limit: Optional[int] = None):
-        """For each PipelineNode, this is run in its own thread."""
-        assert isinstance(self.plugin, (ProcessPlugin, FileInputPlugin))
-
-        logger.info("%s: 0%%", self.name)
-
-        self.plugin.prepare([node.name for node in self.parent_nodes], row_limit)
-
-        if len(self.parent_nodes) == 1:
-            assert isinstance(self.plugin, ProcessPlugin)
-            # there is only a single parent node, run several subthreads to
-            # do the processing
-            only_parent_node = list(self.parent_nodes)[0]
-            only_parent_queue = only_parent_node.add_output_queue()
-            subthreads = [
-                Thread(target=self.run_multithread, args=(only_parent_queue, only_parent_node.name, row_limit))
-                for _ in range(0, 4)
-            ]
-            for subthread in subthreads:
-                subthread.start()
-            for subthread in subthreads:
-                subthread.join()
-
-            self.queue_output(self.plugin.finished(only_parent_node.name))
-
-        elif len(self.parent_nodes) > 1:
-            assert isinstance(self.plugin, ProcessPlugin)
-            # there are multiple parent nodes: spawn off a subthread to handle
-            # each of them.
-            subthreads = [
-                Thread(
-                    target=self.run_subthread,
-                    args=(parent_node.add_output_queue(), parent_node.name, row_limit),
-                )
-                for parent_node in self.parent_nodes
-            ]
-            for subthread in subthreads:
-                subthread.start()
-            for subthread in subthreads:
-                subthread.join()
-
-        self.queue_output(self.plugin.finalize())
-        self.finish_output()
-
-        logger.info("%s: 100%%", self.name)
-
     def load_config(self):
         assert isinstance(self.plugin, BasePlugin)
         if self.config:
@@ -202,33 +87,133 @@ class PipelineNode:
                     logger.warning("Parameter %s=%s Not Valid", key, val)
             self.config = None
 
-    def prerun(self, row_limit=PRERUN_ROW_LIMIT):
-        if not self.plugin:
-            return
-        self.load_config()
-        if self.is_dirty:
-            assert isinstance(self.plugin, (ProcessPlugin, FileInputPlugin))
-            self.result = []
-            self.plugin.prepare([node.name for node in self.parent_nodes], row_limit)
+    cursor = None
+    thread = None
+    status: PipelineNodeStatus = PipelineNodeStatus.INIT
+    table_name = None
 
-            for parent_node in self.parent_nodes:
-                assert isinstance(self.plugin, ProcessPlugin)
-                parent_node.prerun(row_limit)
-                if parent_node.result:
-                    for data_in in parent_node.result:
-                        self.plugin.preprocess(data_in, parent_node.name)
-                    for data_in in parent_node.result:
-                        for data_out in self.plugin.process(data_in, parent_node.name):
-                            self.result.append(data_out)
-                            logger.info("%s: %s/0", self.name, len(self.result))
-                for data_out in self.plugin.finished(parent_node.name):
-                    self.result.append(data_out)
-                    logger.info("%s: %s/0", self.name, len(self.result))
-            for data_out in self.plugin.finalize():
-                self.result.append(data_out)
-                logger.info("%s: %s/0", self.name, len(self.result))
-            self.is_dirty = False
+    def start(self, ddbc, row_limit: Optional[int] = None):
+        """Start processing this operation, in its own thread"""
+        logger.debug("PipelineNode.start %s", self.name)
+
+        if self.table_name:
+            ddbc.sql(f"DROP TABLE IF EXISTS {self.table_name}")
+
+        # start off any parent nodes which aren't done
+        for pn in self.parent_nodes:
+            if pn.status != PipelineNodeStatus.DONE:
+                pn.start(ddbc, row_limit)
+
+        self.load_config()
+
+        # cursor is specific to this thread *for running queries* but is also used from
+        # outside this thread *for interrupting and monitoring the running query*.
+        self.cursor = ddbc.cursor()
+        self.cursor.sql("set enable_progress_bar_print=false")
+        self.cursor.sql("set progress_bar_time=0")
+
+        def _run():
+            logger.debug("PipelineNode.start _run wait %s", self.name)
+
+            # Poll waiting for all parent nodes to be done, or our status
+            # to get changed.
+            self.status = PipelineNodeStatus.WAIT
+            while not all(pn.status == PipelineNodeStatus.DONE for pn in self.parent_nodes):
+                time.sleep(0.1)
+                if self.status != PipelineNodeStatus.WAIT:
+                    return
+
+            self.status = PipelineNodeStatus.WORK
+            logger.debug("PipelineNode.start _run work %s", self.name)
+
+            sources = {pn.name: self.cursor.table(pn.table_name) for pn in self.parent_nodes if pn.table_name}
+            logger.debug("PipelineNode.start _run sources %s", sources.keys())
+            self.plugin.prepare_multi(ddbc, sources)
+            try:
+                result = self.plugin.execute_multi(self.cursor, sources, row_limit)
+                if result is None:
+                    self.table_name = None
+                else:
+                    self.table_name = f"n_{self.uuid}"
+                    result.to_table(self.table_name)
+                self.status = PipelineNodeStatus.DONE
+                logger.debug("PipelineNode.start _run done %s %s", self.name, self.table_name)
+
+            except (duckdb.CatalogException, duckdb.InterruptException) as exc:
+                logger.warning(exc)
+                self.status = PipelineNodeStatus.STOP
+                logger.debug("PipelineNode.start _run stop %s", self.name)
+                self.table_name = None
+
+        # kick off a thread for this process
+        self.thread = threading.Thread(target=_run)
+        self.thread.start()
+
+    def poll(self):
+        for pn in self.parent_nodes:
+            pn.poll()
+
+        if self.status == PipelineNodeStatus.WAIT:
+            logger.info("%s: 0/0", self.name)
+            return True
+
+        if self.status == PipelineNodeStatus.WORK:
+            qp = self.plugin.query_progress(self.cursor)
+            if qp > 0:
+                logger.info("%s: %d%%", self.name, qp)
+            else:
+                logger.info("%s: 0/0", self.name)
+            return True
+
+        if self.status == PipelineNodeStatus.DONE:
             logger.info("%s: 100%%", self.name)
+        else:
+            logger.info("%s: 0/0", self.name)
+        return False
+
+    def stop(self):
+        """Stop any running operation."""
+
+        # interrupt the thread if it is waiting
+        self.status = PipelineNodeStatus.STOP
+
+        # interrupt any running duckdb query
+        if self.cursor:
+            self.cursor.interrupt()
+
+        # wait for the thread to finish
+        if self.thread:
+            self.thread.join()
+
+    def wait(self):
+        if self.status == PipelineNodeStatus.WORK and self.thread:
+            self.thread.join()
+
+    def run(self, ddbc, row_limit: Optional[int] = None):
+        if not self.plugin:
+            return None
+        self.load_config()
+
+        assert isinstance(self.plugin, DuckdbPlugin)
+        if self.is_dirty:
+            sources = {pn.name: pn.run(ddbc, row_limit) for pn in self.parent_nodes}
+            self.plugin.prepare_multi(ddbc, sources)
+            result = self.plugin.execute_multi(ddbc, sources, row_limit)
+            if result is not None:
+                try:
+                    table_name = f"n_{self.uuid}"
+                    ddbc.sql(f"DROP TABLE IF EXISTS {table_name}")
+                    result.to_table(table_name)
+                    self.result = ddbc.table(table_name)
+                    logger.debug("PipelineNode.run saved table %s", table_name)
+                except duckdb.CatalogException as exc:
+                    logger.warning(exc)
+                    self.result = None
+            else:
+                self.result = None
+            self.is_dirty = False
+
+        return self.result
 
     def mark_dirty(self):
         self.is_dirty = True
@@ -262,6 +247,8 @@ class PipelineGraph:
     def __init__(self, nodes: Optional[list[PipelineNode]] = None):
         self.plugin_classes = get_plugin_classes()
         self.nodes = nodes or []
+        self.ddbc = duckdb.connect()
+        self.ddbc.sql("SET python_enable_replacements = false")
 
     def reset_node_name(self, node: PipelineNode):
         node_names_seen = set(n.name for n in self.nodes if n != node)
@@ -288,10 +275,10 @@ class PipelineGraph:
 
     def traverse_nodes(self) -> Iterable[PipelineNode]:
         found_nodes = set(node for node in self.nodes if not node.parent_nodes)
-        yield from found_nodes
+        yield from sorted(found_nodes, key=lambda n: n.uuid)
 
         while len(found_nodes) < len(self.nodes):
-            for node in self.nodes:
+            for node in sorted(self.nodes, key=lambda n: n.uuid):
                 if node not in found_nodes and node.parent_nodes.issubset(found_nodes):
                     yield node
                     found_nodes.add(node)
@@ -306,20 +293,29 @@ class PipelineGraph:
                     yield node
                     found_nodes.add(node)
 
+    def run_node(self, node: PipelineNode):
+        return node.run(self.ddbc)
+
     def run(self):
-        threads_and_nodes = []
+        # Unlike 'start', we kick all the nodes off in one thread so that
+        # their results can stay as views rather than tables.  There's a slight
+        # penalty to this as multiple inputs can't run in parallel but it
+        # should reduce memory usage since intermediate steps don't need to
+        # be saved.  If DuckDB decides to start sharing views across cursors
+        # we can improve on this a little.
+        # see https://github.com/duckdb/duckdb/issues/1848
+
         logger.info("Starting")
         start_time = time.time()
-        for node in self.traverse_nodes_backwards():
+        for node in self.traverse_nodes():
             node.load_config()
-            threads_and_nodes.append((Thread(target=node.run_thread), node))
-
-        for thread, _ in threads_and_nodes:
-            thread.start()
-
-        while any(t.is_alive() for t, _ in threads_and_nodes):
-            logger.info("Elapsed time: %d", time.time() - start_time)
-            time.sleep(10)
+            sources = {pn.name: pn.result for pn in node.parent_nodes}
+            node.plugin.prepare_multi(self.ddbc, sources)
+            result = node.plugin.execute_multi(self.ddbc, sources)
+            if result:
+                node.result = duckdb_source_to_view(self.ddbc, result)
+            else:
+                node.result = None
 
         logger.info("Finished, elapsed time: %d", time.time() - start_time)
 
