@@ -1,12 +1,10 @@
 # TK based GUI for CountESS
-import math
+import logging
 import tkinter as tk
 from functools import partial
 from tkinter import ttk
 from typing import Mapping, MutableMapping, Optional
 
-import numpy as np
-
 from ..core.parameters import (
     ArrayParam,
     BaseParam,
@@ -33,9 +31,7 @@ from .widgets import (
     delete_button,
 )
 
-
-def is_nan(v):
-    return v is None or v is np.nan or (isinstance(v, float) and math.isnan(v))
+logger = logging.getLogger(__name__)
 
 
 class ParameterWrapper:
@@ -393,13 +389,19 @@ class ParameterWrapper:
             self.callback(self.parameter)
 
     def value_changed_callback(self, *_):
-        if isinstance(self.parameter, (ChoiceParam, DictChoiceParam)) and self.entry.current() != -1:
-            value = self.parameter.get_values()[self.entry.current()]
-            if value != self.parameter.value:
-                self.parameter.value = value
-                self.callback(self.parameter)
+        # called from self.var.trace("w", self.value_changed_callback)
+
+        if isinstance(self.parameter, (ChoiceParam, DictChoiceParam)):
+            values = self.parameter.get_values()
+            if 0 <= self.entry.current() < len(values):
+                new_value = values[self.entry.current()]
+                if self.parameter.value != new_value:
+                    self.parameter.set_value(new_value)
+                    self.callback(self.parameter)
+            else:
+                self.var.set(self.set_value(self.var.get()))
         else:
-            self.var.set(self.set_value(self.var.get()))
+            self.set_value(self.var.get())
 
     def widget_modified_callback(self, *_):
         # only gets called the *first* time a modification happens, unless
