@@ -1,16 +1,18 @@
+import getopt
 import logging
 import logging.handlers
 import multiprocessing
 import os
+import queue
 import re
 import sys
-import threading
 import tkinter as tk
 from collections.abc import Callable
 from tkinter import messagebox, ttk
-from typing import Optional
+from typing import Optional, Union
 
 import psutil
+from duckdb import DuckDBPyRelation
 
 from countess import VERSION
 from countess.core.config import export_config_graphviz, read_config, write_config
@@ -28,11 +30,28 @@ from countess.gui.widgets import (
     get_icon,
     info_button,
 )
-from countess.utils.pandas import concat_dataframes
+
+preview_row_limit: Optional[int] = 1000000
+
+usage = """usage: countess_gui [--log LEVEL] [INIFILE]
+
+Start the CountESS GUI.
+
+options:
+    --help                         show this message & exit.
+    --version                      show version
+    --preview LIMIT                set preview mode row limit (default: %s)
+    --log LEVEL                    set log level to LEVEL
+    INIFILE                        load configuration file
+""" % (
+    preview_row_limit or "NONE"
+)
+
 
 # import faulthandler
 # faulthandler.enable(all_threads=True)
 
+# logging.basicConfig(format="%(threadName)s:%(message)s")
 logger = logging.getLogger(__name__)
 
 logging_queue: multiprocessing.Queue = multiprocessing.Queue()
@@ -80,8 +99,9 @@ class ConfiguratorWrapper:
     info_toplevel = None
     info_frame = None
 
-    def __init__(self, tk_parent, node, change_callback):
+    def __init__(self, tk_parent, node, ddbc, change_callback):
         self.node = node
+        self.ddbc = ddbc
         self.change_callback = change_callback
 
         self.frame = ResizingFrame(tk_parent, orientation=ResizingFrame.Orientation.VERTICAL, bg="darkgrey")
@@ -184,38 +204,24 @@ class ConfiguratorWrapper:
         self.notes_widget.grid(row=2, columnspan=2, sticky=tk.EW, padx=10, pady=5)
 
     def show_preview_subframe(self):
+        logger.debug("ConfiguratorWrapper.show_preview_subframe")
         if self.preview_subframe:
             self.frame.remove_child(self.preview_subframe).destroy()
 
         if not self.node.plugin.show_preview:
             return
-        elif not self.node.result:
+
+        if isinstance(self.node.result, DuckDBPyRelation):
+            self.preview_subframe = TabularDataFrame(self.frame, cursor="arrow")
+            self.preview_subframe.set_table(self.ddbc, self.node.result)
+            self.preview_subframe.set_sort_order(self.node.sort_column or 0, self.node.sort_descending)
+            self.preview_subframe.set_callback(self.preview_changed_callback)
+        else:
             self.preview_subframe = tk.Frame(self.frame)
-            self.preview_subframe.columnconfigure(0, weight=1)
-            tk.Label(self.preview_subframe, text="no result").grid(sticky=tk.EW)
-        elif all(isinstance(r, (str, bytes)) for r in self.node.result):
-            text_result = "".join(self.node.result)
-            self.preview_subframe = tk.Frame(self.frame)
-            self.preview_subframe.columnconfigure(0, weight=1)
-            self.preview_subframe.rowconfigure(1, weight=1)
-            n_lines = len(text_result.splitlines())
-            tk.Label(self.preview_subframe, text=f"Text Preview {n_lines} Lines").grid(sticky=tk.NSEW)
             text = tk.Text(self.preview_subframe)
-            text.insert("1.0", text_result)
+            text.insert("1.0", str(self.node.result))
             text["state"] = "disabled"
             text.grid(sticky=tk.NSEW)
-        else:
-            try:
-                df = concat_dataframes(self.node.result)
-                self.preview_subframe = TabularDataFrame(self.frame, cursor="arrow")
-                self.preview_subframe.set_dataframe(df)
-                self.preview_subframe.set_sort_order(self.node.sort_column or 0, self.node.sort_descending)
-                self.preview_subframe.set_callback(self.preview_changed_callback)
-            except (TypeError, ValueError):
-                self.preview_subframe = tk.Frame(self.frame)
-                self.preview_subframe.columnconfigure(0, weight=1)
-                tk.Label(self.preview_subframe, text="no result").grid(sticky=tk.EW)
-
         self.frame.add_child(self.preview_subframe)
 
     def preview_changed_callback(self, offset: int, sort_col: int, sort_desc: bool) -> None:
@@ -233,47 +239,45 @@ class ConfiguratorWrapper:
 
     def config_change_callback(self, *_):
         """Called immediately if a change to config has occurred."""
-        logger.debug("config_change_callback")
+        # logger.debug("config_change_callback")
+        self.node.mark_dirty()
 
-        # Leave it a bit to see if the user is still typing, if so cancel.
+        # Leave it a bit (2500ms) to see if the user is still typing, if so cancel
+        # the previous task and make a new task for another 2.5 seconds away ...
         if self.config_change_task:
+            # logger.debug("config_change_callback: cancelling task")
             self.frame.after_cancel(self.config_change_task)
         self.config_change_task = self.frame.after(2500, self.config_change_task_callback)
 
     def config_change_task_callback(self):
-        """Called when the user makes a change then pauses for a bit"""
-        logger.debug("config_change_task_callback")
-
-        # if there's already an update running, delay a bit longer
-        if self.node_update_thread and self.node_update_thread.is_alive():
-            # XXX need some way to tell the thread to stop
-            logger.debug("config_change_task_callback: waiting for thread %s", self.node_update_thread)
-            self.config_change_task = self.frame.after(1000, self.config_change_task_callback)
-            return
-
+        """Called when the user makes a change and had paused for a bit"""
         self.config_change_task = None
-        self.node.mark_dirty()
+        self.node.stop()
+        self.config_change_start()
 
-        self.node_update_thread = threading.Thread(target=self.node.prerun)
-        logger.debug("config_change_task_callback: starting thread %s", self.node_update_thread)
-        self.node_update_thread.start()
+    def choose_plugin(self, plugin_class):
+        self.node.plugin = plugin_class()
+        self.show_config_subframe()
+        self.config_change_start()
 
-        self.config_change_task_callback_2()
+    def config_change_start(self):
+        self.node.is_dirty = True
+        self.node.start(self.ddbc, preview_row_limit)
+        self.config_change_poll_callback()
 
-    def config_change_task_callback_2(self):
-        logger.debug("config_change_task_callback_2")
+    def config_change_poll_callback(self):
+        if self.node.poll():
+            self.frame.after(50, self.config_change_poll_callback)
+        else:
+            self.config_change_poll_done()
 
-        if self.node_update_thread and self.node_update_thread.is_alive():
-            logger.debug("config_change_task_callback_2: waiting for thread %s", self.node_update_thread)
-            self.frame.after(500, self.config_change_task_callback_2)
-            return
+    def config_change_poll_done(self):
+        self.node.wait()
+        if self.node.table_name:
+            self.node.result = self.ddbc.table(self.node.table_name)
+        else:
+            self.node.result = None
 
-        logger.debug("config_change_task_callback_2: joining thread %s", self.node_update_thread)
-        self.node_update_thread.join()
-        self.node_update_thread = None
-
-        # XXX stop the form scrolling away when it is refreshed, by putting
-        # it back where it belongs.
         pos1, pos2 = self.config_scrollbar.get()
         self.show_preview_subframe()
         self.configurator.update()
@@ -281,19 +285,29 @@ class ConfiguratorWrapper:
         self.config_canvas.yview_moveto(pos1)
         self.config_scrollbar.set(pos1, pos2)
 
-    def choose_plugin(self, plugin_class):
-        self.node.plugin = plugin_class()
-        self.node.prerun()
-        self.node.is_dirty = True
-        self.show_config_subframe()
-        self.change_callback(self.node)
-
     def destroy(self):
         if self.config_change_task:
             self.frame.after_cancel(self.config_change_task)
         self.frame.destroy()
 
 
+class SplashWrapper:
+    def __init__(self, tk_parent):
+        self.frame = tk.Frame(tk_parent)
+        self.frame.configure(cursor="heart")
+
+        font = ("TkHeadingFont", 16, "bold")
+        subframe = tk.Frame(self.frame, highlightbackground="black", highlightthickness=1)
+        subframe.grid(padx=10, pady=10)
+        self.frame.columnconfigure(0, weight=1)
+        self.frame.rowconfigure(0, weight=1)
+        tk.Label(subframe, image=get_icon(tk_parent, "countess")).grid(padx=10, pady=10)
+        tk.Label(subframe, text=f"CountESS {VERSION}", font=font).grid(padx=10, pady=10)
+
+    def destroy(self):
+        self.frame.destroy()
+
+
 class ButtonMenu:  # pylint: disable=R0903
     def __init__(self, tk_parent, buttons):
         self.frame = tk.Frame(tk_parent)
@@ -379,7 +393,11 @@ class LoggerFrame(tk.Frame):
         # Start a QueueListener in its own thread,
         # logging_callback gets called for each record
         # received.
-        logging.handlers.QueueListener(logging_queue, _CallbackLoggingHandler(self.logging_callback)).start()
+        # logging.handlers.QueueListener(logging_queue, _CallbackLoggingHandler(self.logging_callback)).start()
+
+        # XXX tkinter doesn't seem happy with updating widgets from threads,
+        # so here's an alternative, we poll from a tkinter event.
+        self.logging_handler = logging.handlers.QueueListener(logging_queue)
 
         self.columnconfigure(0, weight=1)
         self.rowconfigure(0, weight=1)
@@ -388,9 +406,20 @@ class LoggerFrame(tk.Frame):
         self.pbars_done: list[LabeledProgressbar] = []
         self.hide_event()
 
+        self.logging_event()
+
+    def logging_event(self, ev=None):
+        try:
+            for _ in range(0, 10):
+                rec = self.logging_handler.dequeue(block=False)
+                self.logging_callback(rec)
+            self.after(100, self.logging_event)
+        except queue.Empty:
+            self.after(100, self.logging_event)
+
     def logging_callback(self, record: logging.LogRecord) -> None:
         message = record.getMessage()
-        if m := re.match(r"(.*): (\d+)(%|/\d+)", message):
+        if m := re.match(r"(.*): (\d+(?:\.\d+)?)(%|/\d+)", message):
             name, n1, n2 = m.groups()
             try:
                 pbar = self.progress_bars[name]
@@ -399,7 +428,7 @@ class LoggerFrame(tk.Frame):
                 self.progress_bars[name] = pbar
                 pbar.grid(sticky=tk.EW, row=len(self.progress_bars), column=0)
             if n2 == "%":
-                pbar.progress_update(message, int(n1))
+                pbar.progress_update(message, int(float((n1))))
             else:
                 pbar.progress_update(message)
         else:
@@ -430,6 +459,8 @@ class MainWindow:
     def __init__(self, tk_parent: tk.Widget, config_filename: Optional[str] = None):
         self.tk_parent = tk_parent
 
+        logger.debug("MainWindow.__init__")
+
         ButtonMenu(
             tk_parent,
             [
@@ -449,7 +480,7 @@ class MainWindow:
         # | tree_canvas  | config_wrapper.subframe     |
         # |              |                             |
         # |              +-----------------------------+
-        # |              | configw_wrapper.preview     |
+        # |              | config_wrapper.preview      |
         # |              |                             |
         # +--------------+-----------------------------+
         # | logger_subframe                            |
@@ -471,6 +502,10 @@ class MainWindow:
         self.tree_canvas = FlippyCanvas(self.main_subframe, bg="skyblue")
         self.main_subframe.add_child(self.tree_canvas)
 
+        self.config_wrapper = SplashWrapper(self.main_subframe)
+        self.main_subframe.add_child(self.config_wrapper.frame, weight=4)
+
+        logger.debug("MainWindow.__init__(%s)", repr(config_filename))
         if config_filename:
             self.config_load(config_filename)
         else:
@@ -501,7 +536,7 @@ class MainWindow:
             self.graph_wrapper.destroy()
         self.graph = PipelineGraph()
         self.graph_wrapper = GraphWrapper(self.tree_canvas, self.graph, self.node_select)
-        self.graph_wrapper.add_new_node()
+        self.graph_wrapper.add_new_node(select=False)
 
     def config_load(self, filename=None):
         if not filename:
@@ -559,18 +594,15 @@ class MainWindow:
 
     def node_select(self, node):
         if node:
-            new_config_wrapper = ConfiguratorWrapper(self.main_subframe, node, self.node_changed)
-            if self.config_wrapper:
-                self.main_subframe.replace_child(self.config_wrapper.frame, new_config_wrapper.frame)
-                self.config_wrapper.destroy()
-            else:
-                self.main_subframe.add_child(new_config_wrapper.frame, weight=4)
+            new_config_wrapper = ConfiguratorWrapper(self.main_subframe, node, self.graph.ddbc, self.node_changed)
+        else:
+            new_config_wrapper = SplashWrapper(self.main_subframe)
 
-            self.config_wrapper = new_config_wrapper
+        if self.config_wrapper:
+            self.main_subframe.replace_child(self.config_wrapper.frame, new_config_wrapper.frame)
+            self.config_wrapper.destroy()
 
-        elif self.config_wrapper:
-            self.main_subframe.remove_child(self.config_wrapper.frame).destroy()
-            self.config_wrapper = None
+        self.config_wrapper = new_config_wrapper
 
     def node_changed(self, node):
         self.graph_wrapper.node_changed(node)
@@ -633,19 +665,56 @@ def make_root():
     return root
 
 
-def main():
+def main() -> None:
+    args = sys.argv[1:]
+    global preview_row_limit  # pylint: disable=global-statement
+
+    logging.getLogger().setLevel("INFO")
+
+    try:
+        options, args = getopt.getopt(args, "", ["help", "version", "preview=", "log="])
+    except getopt.GetoptError as exc:
+        logger.error(str(exc))
+        sys.exit(1)
+
+    for opt_key, opt_val in options:
+        if opt_key == "--help":
+            print(usage)  # pylint: disable=bad-builtin
+            sys.exit(0)
+        elif opt_key == "--version":
+            print(f"CountESS {VERSION}")  # pylint: disable=bad-builtin
+            sys.exit(0)
+        elif opt_key == "--preview":
+            if opt_val.upper() == "NONE":
+                preview_row_limit = None
+            elif re.match(r"\d+$", opt_val):
+                preview_row_limit = int(opt_val) or None
+            else:
+                logger.warning("Bad --preview value: %s, expected a positive integer or 'NONE'", opt_val)
+        elif opt_key == "--log":
+            try:
+                log_level: Union[int, str]
+                if re.match(r"\d+$", opt_val):
+                    log_level = int(opt_val)
+                else:
+                    log_level = str(opt_val).upper()
+                logging.getLogger().setLevel(log_level)
+                logger.warning("Log level set to %s", log_level)
+            except ValueError:
+                logger.error("Bad --log level: %s", opt_val)
+
+    logger.info("Preview Limit set to %d", preview_row_limit)
+
     # set up a multiprocessing-compatible logging queue to bring all logging
     # messages back to the main process.
     logging.getLogger().addHandler(logging.handlers.QueueHandler(logging_queue))
-    logging.getLogger().addHandler(logging.StreamHandler())
-    logging.getLogger().setLevel(logging.INFO)
+    # logging.getLogger().addHandler(logging.StreamHandler())
 
     root = make_root()
-    SplashScreen(root)
-    MainWindow(root, sys.argv[1] if len(sys.argv) > 1 else None)
+    MainWindow(root, args[0] if args else None)
 
     root.mainloop()
 
 
 if __name__ == "__main__":
-    main()
+    main()  # pragma: no cover
