@@ -1,14 +1,22 @@
 import io
+import logging
 import tkinter as tk
 from functools import partial
 from math import ceil, floor, isinf, isnan
 from tkinter import ttk
 from typing import Callable, Optional, Union
 
-import pandas as pd
-from pandas.api.types import is_integer_dtype, is_numeric_dtype
+from duckdb import DuckDBPyConnection, DuckDBPyRelation
 
 from countess.gui.widgets import ResizingFrame, copy_to_clipboard, get_icon
+from countess.utils.duckdb import (
+    duckdb_dtype_is_integer,
+    duckdb_dtype_is_numeric,
+    duckdb_dtype_to_datatype_choice,
+    duckdb_escape_identifier,
+)
+
+logger = logging.getLogger(__name__)
 
 # XXX columns should automatically resize based on information
 # from _column_xscrollcommand which can tell if they're
@@ -17,18 +25,27 @@ from countess.gui.widgets import ResizingFrame, copy_to_clipboard, get_icon
 # etc etc.
 
 
-def column_format_for(df_column: Union[pd.Index, pd.Series]) -> str:
-    if is_numeric_dtype(df_column.dtype):
+def column_format_for(table: DuckDBPyRelation, column: str) -> str:
+    # logger.debug("column_format_for column %s %s", column, table.columns)
+
+    # XXX https://github.com/duckdb/duckdb/issues/15267
+    # dtype = table[column].dtypes[0]
+    dtype = table.project(duckdb_escape_identifier(column)).dtypes[0]
+
+    if duckdb_dtype_is_numeric(dtype):
         # Work out the maximum width required to represent the integer part in this
         # column, so we can pad values to that width.
-        column_min = df_column.min()
-        if isnan(column_min) or isinf(column_min):
+        column_esc = duckdb_escape_identifier(column)
+        column_min_max = table.aggregate(f"min({column_esc}), max({column_esc})").fetchone()
+        if column_min_max is None:
+            return "%s"
+        column_min, column_max = column_min_max
+        if column_min is None or isnan(column_min) or isinf(column_min):
             column_min = -100
-        column_max = df_column.max()
-        if isnan(column_max) or isinf(column_max):
+        if column_max is None or isnan(column_max) or isinf(column_max):
             column_max = 100
         width = max(len(str(floor(column_min))), len(str(ceil(column_max))))
-        if is_integer_dtype(df_column.dtype):
+        if duckdb_dtype_is_integer(dtype):
             return f"%{width}d"
         else:
             # leave room for the point and 12 decimals.
@@ -70,12 +87,14 @@ def format_value(value: Optional[Union[int, float, str]], column_format: str) ->
 
 
 class TabularDataFrame(tk.Frame):
-    """A frame for displaying a pandas (or similar) dataframe.
+    """A frame for displaying a duckdb relation
     Columns are displayed as individual tk.Text widgets which seems to be relatively efficient
     as they only hold the currently displayed rows.  Tested up to a million or so rows."""
 
     subframe: Optional[tk.Frame] = None
-    dataframe: Optional[pd.DataFrame] = None
+    ddbc: Optional[DuckDBPyConnection] = None
+    table: Optional[DuckDBPyRelation] = None
+    table_order: str = ""
     offset = 0
     height = 1000
     length = 0
@@ -84,14 +103,14 @@ class TabularDataFrame(tk.Frame):
     columns: list[tk.Text] = []
     column_formats: list[str] = []
     scrollbar = None
-    index_cols = 0
-    sort_by_col = None
+    sort_by_col: Optional[int] = None
     sort_ascending = True
     callback: Optional[Callable[[int, int, bool], None]] = None
     click_callback: Optional[Callable[[int, int, int], None]] = None
 
     def __init__(self, *a, **k):
         super().__init__(*a, **k)
+
         self.rowconfigure(0, weight=0)
         self.rowconfigure(1, weight=0)
         self.rowconfigure(2, weight=1)
@@ -110,58 +129,29 @@ class TabularDataFrame(tk.Frame):
 
         self.bind("<Configure>", self._configure)
 
-    def set_dataframe(self, dataframe: pd.DataFrame, offset: Optional[int] = 0):
-        self.dataframe = dataframe
-        self.length = len(dataframe)
+    def set_table(self, ddbc: DuckDBPyConnection, table: DuckDBPyRelation, offset: Optional[int] = 0):
+        self.ddbc = ddbc
+        self.table = table
+        self.length = len(table)
 
-        # clean up column names
-
-        if hasattr(self.dataframe.index, "names") and hasattr(self.dataframe.index, "dtypes"):
-            # MultiIndex case
-            column_names = list(self.dataframe.index.names) + list(self.dataframe.columns)
-            column_dtypes = list(self.dataframe.index.dtypes) + list(self.dataframe.dtypes)
-            index_frame = self.dataframe.index.to_frame()
-            self.column_formats = [column_format_for(index_frame[name]) for name in dataframe.index.names] + [
-                column_format_for(dataframe[name]) for name in dataframe.columns
-            ]
-            self.index_cols = len(self.dataframe.index.names)
-        elif self.dataframe.index.name:
-            # a simple Index, with a name
-            column_names = [self.dataframe.index.name] + list(self.dataframe.columns)
-            column_dtypes = [self.dataframe.index.dtype] + list(self.dataframe.dtypes)
-            self.column_formats = [column_format_for(dataframe.index)] + [
-                column_format_for(dataframe[name]) for name in dataframe.columns
-            ]
-            self.index_cols = 1
-        else:
-            # if it doesn't have a name, don't bother displaying it
-            # XXX it's probably just a RangeIndex, should we display it anyway?
-            column_names = list(self.dataframe.columns)
-            column_dtypes = list(self.dataframe.dtypes)
-            self.column_formats = [column_format_for(dataframe[name]) for name in dataframe.columns]
-            self.index_cols = 0
+        column_names = table.columns
+        column_dtypes = table.dtypes
+        self.column_formats = [column_format_for(table, name) for name in column_names]
 
         n_columns = len(column_names)
         if n_columns == 0:
             self.label["text"] = "Dataframe Preview\n\nno data"
             return
 
-        # make labels for all the columns
-        ### XXX add in proper handling for MultiIndexes here
-
+        # make new labels for all the columns
         for label in self.labels:
             label.destroy()
-
         self.labels = []
         for num, (name, dtype) in enumerate(zip(column_names, column_dtypes)):
-            if type(name) is tuple:
-                name = "\n".join([str(n) for n in name])
-            else:
-                name = str(name)
-            is_index = " (index)" if num < self.index_cols else ""
+            label_text = f"{name}\n{duckdb_dtype_to_datatype_choice(dtype)}"
             column_label = tk.Label(
                 self.label_frame,
-                text=f"{name}\n{dtype}{is_index}",
+                text=label_text,
                 image=get_icon(self, "sort_un"),
                 compound=tk.RIGHT,
             )
@@ -175,7 +165,7 @@ class TabularDataFrame(tk.Frame):
         self.subframe = ResizingFrame(self, orientation=ResizingFrame.Orientation.HORIZONTAL, bg="darkgrey")
         self.subframe.pack(side="left", fill="both", expand=True)
 
-        self.label["text"] = f"Dataframe Preview {len(self.dataframe)} rows"
+        self.label["text"] = f"Dataframe Preview {len(self.table)} rows"
 
         self.columns = []
         for num, (name, dtype) in enumerate(zip(column_names, column_dtypes)):
@@ -192,73 +182,34 @@ class TabularDataFrame(tk.Frame):
             column_text.bind("<Configure>", partial(self._column_configure, num))
             self.columns.append(column_text)
 
-    def _column_configure(self, num, ev):
+    def _column_configure(self, num: int, ev):
         # when the column changes position, move the labels around
         # to match
         self.labels[num].place(x=ev.x, width=ev.width)
 
-    def _column_click(self, col, ev):
+    def _column_click(self, col: int, ev):
         if self.click_callback:
             line, char = ev.widget.index("current").split(".")
             row = min(self.offset + int(line), self.length) - 1
             self.click_callback(col, row, int(char))
 
-    def refresh(self, new_offset=0):
+    def refresh(self, new_offset: float = 0):
         # Refreshes the column widgets.
-        # XXX should handle new_height as well, as this changes a fair bit
-        # with some window managers. Needs refactoring.
+        assert self.table is not None
+        assert self.scrollbar
 
-        new_offset = max(0, min(self.length - self.height, int(new_offset)))
-        offset_diff = new_offset - self.offset
+        self.offset = max(0, min(self.length - self.height, int(new_offset)))
 
-        # get the new rows as an iterator
-        if 1 <= offset_diff < self.height:
-            df = self.dataframe.iloc[self.offset + self.height : self.offset + self.height + offset_diff]
-            insert_at = tk.END
-        elif 1 <= -offset_diff < self.height:
-            # Get rows in reverse order so they can be inserted at the start
-            # note offset_diff is negative!
-            # XXX check this isn't horribly inefficient with pandas indexes
-            if new_offset:
-                df = self.dataframe.iloc[self.offset - 1 : new_offset - 1 : -1]
-            else:
-                df = self.dataframe.iloc[self.offset - 1 :: -1]
-            insert_at = "1.0"
-        else:
-            df = self.dataframe.iloc[new_offset : new_offset + self.height + 1]
-            insert_at = tk.END
+        table = self.table.order(self.table_order) if self.table_order else self.table
+        rows = table.limit(self.height, offset=self.offset).fetchall()
 
-        # then unlock the columns and delete unnecessary rows
-        for cw in self.columns:
-            cw["state"] = tk.NORMAL
-            if 1 <= offset_diff < self.height:
-                # delete rows at start, add new rows on the end
-                cw.delete("1.0", f"{offset_diff+1}.0")
-            elif 1 <= -offset_diff < self.height:
-                # delete rows at end, insert new rows at start
-                # note offset_diff is negative!  Note we have to
-                # restore the deleted final "\n".  Sigh.
-                cw.delete(f"{self.height+offset_diff+2}.0", tk.END)
-                cw.insert(tk.END, "\n")
-            else:
-                # delete everything & add all new rows
-                cw.delete("1.0", tk.END)
+        for column_num, (column_widget, column_format) in enumerate(zip(self.columns, self.column_formats)):
+            column_widget["state"] = tk.NORMAL
+            column_widget.delete("1.0", tk.END)
+            for row in rows:
+                column_widget.insert(tk.END, format_value(row[column_num], column_format) + "\n")
+            column_widget["state"] = tk.DISABLED
 
-        for idx, row in df.iterrows():
-            if type(idx) is tuple:
-                values = list(idx) + list(row)
-            elif self.dataframe.index.name:
-                values = [idx] + list(row)
-            else:
-                values = list(row)
-
-            for value, column, column_format in zip(values, self.columns, self.column_formats):
-                column.insert(insert_at, format_value(value, column_format) + "\n")
-
-        for cw in self.columns:
-            cw["state"] = tk.DISABLED
-
-        self.offset = new_offset
         if self.length:
             self.scrollbar.set(self.offset / self.length, (self.offset + self.height) / self.length)
 
@@ -271,20 +222,46 @@ class TabularDataFrame(tk.Frame):
     def set_click_callback(self, click_callback) -> None:
         self.click_callback = click_callback
 
+    _set_sort_order_thread = None
+
     def set_sort_order(self, column_num: int, descending: Optional[bool] = None):
-        assert self.dataframe is not None
+        assert self.ddbc is not None
+        assert self.table is not None
+
+        if not 0 <= column_num < len(self.table.columns):
+            column_num = 0
 
         if descending is None and column_num == self.sort_by_col:
             self.sort_ascending = not self.sort_ascending
         else:
-            self.sort_by_col = column_num
             self.sort_ascending = not descending
-        if column_num < self.index_cols:
-            self.dataframe = self.dataframe.sort_index(level=column_num, ascending=self.sort_ascending)
-        elif column_num < self.index_cols + len(self.dataframe.columns):
-            self.dataframe = self.dataframe.sort_values(
-                self.dataframe.columns[column_num - self.index_cols], ascending=self.sort_ascending
-            )
+        self.sort_by_col = column_num
+
+        # Create an index for whatever column we're sorting by, followed by
+        # all other column names for a stable sort order.
+
+        column_name = self.table.columns[column_num]
+        escaped_column_names = [
+            duckdb_escape_identifier(c) for c in [column_name] + [c for c in self.table.columns if c != column_name]
+        ]
+
+        index_identifier = f"{self.table.alias}_{column_num}_idx"
+
+        # XXX I did have some code here which built the index in a separate thread / cursor but it
+        # was mad with race conditions so I've scrapped it for now and the GUI will just have to
+        # awkwardly pause.
+
+        sql = (
+            f"CREATE INDEX IF NOT EXISTS {index_identifier} ON {self.table.alias} ("
+            + ",".join(escaped_column_names)
+            + ")"
+        )
+        self.ddbc.sql(sql)
+
+        direction = "ASC" if self.sort_ascending else "DESC"
+        self.table_order = ",".join(f"{c} {direction}" for c in escaped_column_names)
+
+        logger.debug("TabularDataFrame set_sort_order %s %s", column_name, self.table_order)
 
         for n, label in enumerate(self.labels):
             icon = "sort_un" if n != column_num else "sort_up" if self.sort_ascending else "sort_dn"
@@ -370,11 +347,15 @@ class TabularDataFrame(tk.Frame):
         if not self.select_rows:
             return  # not multi-row, keep it.
 
-        # Dump TSV into a StringIO ...
+        # pick the selected rows
+        table = self.table.order(self.table_order) if self.table_order else self.table
         r1, r2 = self.select_rows
-        df = self.dataframe.iloc[self.offset + r1 - 1 : self.offset + r2]
-        buf = io.StringIO()
-        df.to_csv(buf, sep="\t")
+        limit = r2 - r1 + 1
+        offset = self.offset + r1 - 1
+        logger.debug("TabularDataFrame._column_copy r1 %d r2 %d limit %d offset %d", r1, r2, limit, offset)
+        table = table.limit(limit, offset=offset)
 
-        # ... and then push that onto the clipboard
+        # Dump TSV into a StringIO and push it onto the clipboard
+        buf = io.StringIO()
+        table.to_df().to_csv(buf, sep="\t", index=False)
         copy_to_clipboard(buf.getvalue())
