@@ -1,26 +1,22 @@
-from typing import Iterable
+import logging
+from typing import Optional
 
-import pandas as pd
+from duckdb import DuckDBPyConnection, DuckDBPyRelation
 
 from countess import VERSION
-from countess.core.parameters import (
-    ArrayParam,
-    BooleanParam,
-    DataTypeChoiceParam,
-    MultiParam,
-    StringParam,
-    TabularMultiParam,
-)
-from countess.core.plugins import PandasInputPlugin
+from countess.core.parameters import ArrayParam, DataTypeChoiceParam, MultiParam, StringParam, TabularMultiParam
+from countess.core.plugins import DuckdbInputPlugin
+from countess.utils.duckdb import duckdb_escape_identifier, duckdb_escape_literal
+
+logger = logging.getLogger(__name__)
 
 
 class _ColumnsMultiParam(MultiParam):
     name = StringParam("Name")
-    type = DataTypeChoiceParam("Type", "string")
-    index = BooleanParam("Index?")
+    type = DataTypeChoiceParam("Type")
 
 
-class DataTablePlugin(PandasInputPlugin):
+class DataTablePlugin(DuckdbInputPlugin):
     """DataTable"""
 
     name = "DataTable"
@@ -61,17 +57,20 @@ class DataTablePlugin(PandasInputPlugin):
             self.fix_columns()
         super().set_parameter(key, *a, **k)
 
-    def finalize(self) -> Iterable[pd.DataFrame]:
+    def execute(
+        self, ddbc: DuckDBPyConnection, source: None, row_limit: Optional[int] = None
+    ) -> Optional[DuckDBPyRelation]:
         self.fix_columns()
-        values = []
-        for row in self.rows:
-            values.append({str(col.name): row[str(col.name)].value for col in self.columns})
 
-        df = pd.DataFrame(values)
+        if len(self.rows) == 0:
+            return None
 
-        index_cols = [str(col.name) for col in self.columns if col.index]
+        values_str = ", ".join(
+            "(" + (", ".join(duckdb_escape_literal(val.value) for val in row.values())) + ")" for row in self.rows
+        )
+        columns_str = ",".join(duckdb_escape_identifier(col.name.value) for col in self.columns)
+        sql = f"SELECT * FROM (VALUES {values_str}) _({columns_str})"
 
-        if index_cols:
-            df = df.set_index(index_cols)
+        logger.debug("DataTablePlugin.execute sql %s", sql)
 
-        yield df
+        return ddbc.sql(sql)
