@@ -224,7 +224,7 @@ def find_variant_dna(ref_seq: str, var_seq: str, offset: int = 0) -> Iterable[st
     if not re.match("[AGTCN]+$", ref_seq):
         raise ValueError("Invalid reference sequence")
 
-    if not re.match("[AGTCN]+$", var_seq):
+    if not re.match("[AGTC]+$", var_seq):
         raise ValueError("Invalid variant sequence")
 
     # Levenshtein algorithm finds the overlapping parts of our reference and
@@ -395,10 +395,10 @@ def find_variant_protein(ref_seq: str, var_seq: str, offset: int = 0):
     ref_seq = ref_seq.strip().upper()
     var_seq = var_seq.strip().upper()
 
-    if not re.match("[AGTCN]+$", ref_seq):
+    if not re.match("[AGTC]+$", ref_seq):
         raise ValueError("Invalid reference sequence")  # pragma: no cover
 
-    if not re.match("[AGTCN]+$", var_seq):
+    if not re.match("[AGTC]+$", var_seq):
         raise ValueError("Invalid variant sequence")  # pragma: no cover
 
     frame = (3 - offset) % 3
@@ -536,11 +536,6 @@ def find_variant_string(
 
     CHECK FOR INVALID INPUTS
 
-    >>> find_variant_string("x.", "CAT", "CAT")
-    Traceback (most recent call last):
-     ...
-    ValueError: Only prefix types 'g.', 'c.' and 'p.' accepted at this time
-
     >>> find_variant_string("g.", "HELLO", "CAT")
     Traceback (most recent call last):
      ...
@@ -563,12 +558,10 @@ def find_variant_string(
         ref_seq = reverse_complement(ref_seq)
         var_seq = reverse_complement(var_seq)
 
-    if prefix.endswith("g.") or prefix.endswith("c."):
-        variations = list(find_variant_dna(ref_seq, var_seq, offset))
-    elif prefix.endswith("p."):
+    if prefix.endswith("p."):
         variations = list(find_variant_protein(ref_seq, var_seq, offset))
     else:
-        raise ValueError("Only prefix types 'g.', 'c.' and 'p.' accepted at this time")
+        variations = list(find_variant_dna(ref_seq, var_seq, offset))
 
     if len(variations) == 0:
         return prefix + "="
